package uulm.teamname.marvelous.server.lobby;

import org.tinylog.Logger;
import uulm.teamname.marvelous.gamelibrary.entities.EntityType;
import uulm.teamname.marvelous.gamelibrary.events.Event;
import uulm.teamname.marvelous.gamelibrary.events.EventBuilder;
import uulm.teamname.marvelous.gamelibrary.events.EventType;
import uulm.teamname.marvelous.gamelibrary.gamelogic.GameInstance;
import uulm.teamname.marvelous.gamelibrary.messages.ParticipantType;
import uulm.teamname.marvelous.gamelibrary.config.CharacterConfig;
import uulm.teamname.marvelous.gamelibrary.config.PartyConfig;
import uulm.teamname.marvelous.gamelibrary.config.ScenarioConfig;
import uulm.teamname.marvelous.gamelibrary.requests.Request;
import uulm.teamname.marvelous.server.lobby.pipelining.*;
import uulm.teamname.marvelous.server.lobbymanager.LobbyConnection;
import uulm.teamname.marvelous.server.lobbymanager.Participant;

import java.util.*;

public class Lobby {
    public final String gameID;
    public final LobbyConnection connection;
    public final GameInstance game;
    public final Pipeline pipeline;
    private Participant activePlayer;
    private int badRequests;
    private PauseSegment pauseSegment;
    private final TurnTimer turnTimer;

    public Lobby(
            String gameID,
            LobbyConnection connection,
            PartyConfig partyConfig,
            CharacterConfig characterConfig,
            ScenarioConfig scenarioConfig) {
        this.gameID = gameID;
        this.connection = connection;

        //partyConfig.maxRoundTime;

        this.game = new GameInstance(
                partyConfig,
                characterConfig,
                scenarioConfig
        );

        this.pipeline = new Pipeline();

        var reqSegment = new RequestGameStateSegment(this.game);
        this.pauseSegment = new PauseSegment();
        var disconnectSegment = new DisconnectSegment(this);
        var gameStateSegment = new GameStateSegment(this.game);

        pipeline.addSegment(reqSegment)
                .addSegment(pauseSegment)
                .addSegment(disconnectSegment)
                .addSegment(gameStateSegment);

        this.turnTimer = new TurnTimer(
                partyConfig.maxRoundTime,
                this::turnTimeout);
    }

    /**
     * Called by {@link LobbyConnection} to handle requests
     * @param requests to be processed
     * @param source the player executing the requests
     */
    public synchronized void receiveRequests(Request[] requests, Participant source) {
        if(activePlayer != source && source.type != ParticipantType.Spectator) {
            activePlayer = source;
            badRequests = 0;
        }

        Logger.info("got {} requests from participant {}",
                requests.length,
                source);

        Logger.trace("Processing requests through pipeline");
        Optional<Event[]> resultingEvents = pipeline.processRequests(requests, source);
        Logger.debug("generated {} events from the pipeline", resultingEvents.map(x -> x.length).orElse(0));

        //resultingEvents isEmpty when a wrong request appeared
        if (resultingEvents.isEmpty()) {
            reject(source);
        } else {
            var events = resultingEvents.get();
            connection.sendEvents(source, events);
            badRequests = 0;
        }


        EntityType currentActiveCharacterType = game.state.getActiveCharacter().type;
        if (pauseSegment.isPaused()) {
            turnTimer.clear();
        } else if (currentActiveCharacterType == EntityType.P1) {
            turnTimer.startTurnTimer(connection.getPlayer1());
        } else if (currentActiveCharacterType == EntityType.P2) {
            turnTimer.startTurnTimer(connection.getPlayer2());
        } else {
            turnTimer.clear();
        }
    }

    /**
     * If the player executed a false request the request gets rejected.
     * @param source the executing player
     */
    private synchronized void reject(Participant source) {
        connection.sendEvents(source, new EventBuilder(EventType.Nack).buildGameEvent(), game.getGameStateEvent());
        badRequests ++;
        //if the player sends 2 bad messages after one another, the player gets kicked out of the lobby.
        if(badRequests == 2){
            connection.removePlayer(source);
            if(connection.hasPlayer1()){
                    generateWin(connection.getPlayer1());
                }
            else if(connection.hasPlayer2()) {
                    generateWin(connection.getPlayer2());
            }
        }
    }

    /**
     * Warns the player he get timeouted soon.
     * @param source soon to be timeouted player
     */
    public synchronized void soonTimeout (Participant source){
        connection.sendEvents(
                source,
                new EventBuilder(EventType.TimeoutWarningEvent).buildGameStateEvent());
    }

    /**
     * If a player times out the other player automatically wins.
     * @param source is the timeouted player
     */
    public synchronized void timeout(Participant source){
        connection.sendEvents(
                source,
                new EventBuilder(EventType.TimeoutEvent).buildGameStateEvent());
        connection.removePlayer(source);

        if(connection.hasPlayer1()){
            generateWin(connection.getPlayer1());
        }
        else if(connection.hasPlayer2()) {
            generateWin(connection.getPlayer2());
        }
        else {
            throw new IllegalStateException("Spectator was time-outed which is impossible");
        }
    }

    public synchronized void turnTimeout(Participant source){
        //source round is over next players turn
        connection.sendEvents(source, game.endTurn().toArray(new Event[0]));
    }

    /**
     * The method generates a Win event for a {@link Participant}.
     * Afterwards it sends a Disconnect to everyone,  and terminates the connection.
     *
     * @param winner is the {@link Participant} that won
     */
    public synchronized void generateWin(Participant winner){
        connection.broadcastEvents(
                new EventBuilder(EventType.WinEvent)
                        .withPlayerWon(winner.type.equals(ParticipantType.PlayerOne) ? 1 : 2)
                        .buildGameStateEvent(),
                new EventBuilder(EventType.DisconnectEvent)
                        .buildGameStateEvent());
        connection.terminateConnection();
    }

    public PauseSegment getPauseSegment() {
        return pauseSegment;
    }
}
