package uulm.teamname.marvelous.server.lobbymanager;

import uulm.teamname.marvelous.gamelibrary.events.Event;
import uulm.teamname.marvelous.gamelibrary.messages.ParticipantType;
import uulm.teamname.marvelous.server.lobby.Lobby;

import java.util.HashSet;

/**
 * A class that handles the connection to the lobby. It contains distinct websockets for player1, player2 and spectators.
 * The class is meant to be used in conjecture with {@link MessageRelay}.
 */
public class LobbyConnection {
    private Lobby lobby;
    private Participant player1, player2;
    private final HashSet<Participant> spectators;

    /**
     * Creates a new LobbyConnection from a given lobby
     * @param lobby is the lobby that will be connected to
     */
    public LobbyConnection(Lobby lobby) {
        this.lobby = lobby;
        this.spectators = new HashSet<>(10);
    }

    /**
     * @return whether there is a player1
     */
    public boolean hasPlayer1() {
        return player1 != null;
    }

    /**
     * @return whether there is a player2
     */
    public boolean hasPlayer2() {
        return player2 != null;
    }

    public Participant getPlayer1() {
        return player1;
    }

    public Participant getPlayer2() {
        return player2;
    }

    /**
     * Adds a new player into the player1 slot
     * @param player is the websocket to be added
     * @return true if added successfully, and false otherwise
     */
    public boolean addPlayer1(Participant player) {
        if (this.contains(player)) return false;
        if (player1 == null) {
            player1 = player;
            return true;
        } else {
            return false;
        }
    }

    /**
     * Adds a new player into the player1 slot
     * @param player is the websocket to be added
     * @return true if added successfully, and false otherwise
     */
    public boolean addPlayer2(Participant player) {
        if (this.contains(player)) return false;
        if (player2 == null) {
            player2 = player;
            return true;
        } else {
            return false;
        }
    }

    /**
     * Adds a new player into either the player1, or if full, player2 slot
     * @param player is the websocket to be added
     * @return true if added successfully, and false otherwise
     */
    public boolean addPlayer(Participant player) {
        if (!addPlayer1(player)) {
            return addPlayer2(player);
        }
        return true;
    }

    public boolean removePlayer(Participant player) {
        if (player1 == player) {
            player1 = null;
            return true;
        } else if (player2 == player) {
            player2 = null;
            return true;
        } else {
            return spectators.remove(player);
        }
    }


    public boolean contains(Participant participant) {
        return player1 == participant || player2 == participant || spectators.contains(participant);
    }

    // Methods to send events

    public void sendEvents(Participant recipient, Event... events) {
        MessageRelay.getInstance().sendMessage(this, recipient, events);
    }

    public void broadcastEvents(Event... events) {
        // TODO: implement
        MessageRelay.getInstance().broadcastEvents(this, events);
    }

    public void broadcastToAllExcept(Participant except, Event... events) {
        // TODO: implement
        var messageRelayInstance = MessageRelay.getInstance();
        if (except.type == ParticipantType.Spectator) {
            spectators.stream()
                    .filter(spectator -> !spectator.equals(except))
                    .forEach(spectator -> messageRelayInstance.sendMessage(this, spectator, events));
            messageRelayInstance.sendMessage(this, player1, events);
            messageRelayInstance.sendMessage(this, player2, events);
        } else {
            messageRelayInstance.sendMessage(this, except.equals(player1) ? player2 : player1, events);
            spectators.forEach(spectator -> messageRelayInstance.sendMessage(this, spectator, events));
        }
    }

    /** Kills all connections to client, as well as the lobby */
    public void terminateConnection() {
        MessageRelay.getInstance().terminate(this);
    }
}
